import os
import functools
import time
from .test_jcl import TestJcl
from .test_case import TestCase
from .exceptions import ZoweConfigurationNotFound
from .background_colors import BG
from .logger import LOG
from decouple import config
from decouple import UndefinedValueError
from zowe.zos_files_for_zowe_sdk import Files




try:
    ZOWE_CONNECTION = {'plugin_profile': config('ZOWE_ZOSMF_PROFILE')}
except UndefinedValueError:
    try:
        host_url = config('ZOWE_ZOSMF_URL')
        user = config('ZOWE_ZOSMF_USER')
        password = config('ZOWE_ZOSMF_PASSWORD')
        if host_url and user and password:
            ZOWE_CONNECTION = {
                'host_url': host_url,
                'user': user,
                'password': password,
                'ssl_verification': False
            }
    except UndefinedValueError:
        raise ZoweConfigurationNotFound('Missing required environment variable for ZOWE z/OSMF \
                         [ZOWE_ZOSMF_URL,ZOWE_ZOSMF_USER,ZOWE_ZOSMF_PASSWORD] or ZOWE_ZOSMF_PROFILE')

ZOS_FILES = Files(ZOWE_CONNECTION)


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        test_file_object = args[0]
        setattr(test_file_object, '_run_time', run_time)
        return value
    return wrapper_timer

class TestFile:

    def __init__(self, file_object):
        self.file_object = file_object
        self.code = self.file_object.code
        self.test_definitions = []
        self.test_dataset = config('TREXX_TEST_DATASET')
        self.member_name = 'TEMPREXX'

    @timer
    def run(self):
        self.__setup()
        self.__trigger_test_rexx()
        self.__validate_results()
        self.__teardown()

    def __setup(self):
        LOG.debug(f'Initiating test file "{self.file_object.file_name}" setup process')
        self.__get_test_definitions()
        self.__upload_test_file()

    def __get_test_definitions(self):
        LOG.debug(f'Retrieving definitions from test file "{self.file_object.file_name}"')
        test_id_number = 0
        for docstring in self.code.docstring_statements:
            if docstring.statement == 'TEST':
                LOG.debug(f'Test definition found "{docstring.values}"')
                test_id_number += 1
                test_id = f'TEST{test_id_number}'
                self.test_definitions.append(TestCase(test_id,
                                                      docstring.values))

    def __upload_test_file(self):
        LOG.debug(f'Uploading test rexx file to "{self.test_dataset}({self.member_name})"')
        ZOS_FILES.upload_file_to_dsn(self.file_object.file_path,
                                     f'{self.test_dataset}({self.member_name})')

    def __trigger_test_rexx(self):
        LOG.debug('Executing test rexx file...')
        self.test_jcl = TestJcl(self.test_dataset, ZOWE_CONNECTION)
        self.test_jcl.trigger_rexx(self.member_name)
        self.execution_results = self.test_jcl.result_records

    def __validate_results(self):
        LOG.debug(f'Initiating validation process')
        self.raw_test_results = []
        for record in self.execution_results:
            record = record.strip()
            if record.startswith('%TEST'):
                record_columns = record.split('=')
                test_id = record_columns[0][1:]
                test_result = record_columns[1].split(',')[0]
                self.raw_test_results.append({
                    'test_id': test_id,
                    'test_result': test_result,
                    'raw': record
                })

        if len(self.test_definitions) == len(self.raw_test_results):
            for test_definition in self.test_definitions:
                for test_result in self.raw_test_results:
                    if test_definition.test_id == test_result['test_id']:
                        if test_result['test_result'] == 'SUCCESS':
                            test_definition.passed = True
                        test_definition.raw_test_result = test_result['raw']
        else:
            LOG.error(f'Missing @TEST docstring statement in test file "{BG.FAIL}{self.file_object.file_path}{BG.ENDC}", unable to display results')
            exit(1)

    def __teardown(self):
        LOG.debug("Cleaning up")
        self.test_jcl.delete_from_spool()



from .background_colors import BG

class TestCase:

    def __init__(self, test_id, description):
        self.test_id = test_id
        self.description = description
        self.passed = False
        self.raw_test_result = ''

    def __doc__(self):
        return f'{self.test_id} - {self.description}'

    def display_test_results(self, verbose=False):
        self.__set_result_character()
        result_string = f'{self.result_char} {self.__doc__()}'
        suffix = f'\n  ├⎯⎯⎯⎯ {self.raw_test_result}\n'
        if verbose or not self.passed:
            result_string += suffix
        print(result_string)

    def __set_result_character(self):
        if self.passed:
            self.result_char = f'{BG.OKGREEN}✓{BG.ENDC}'
        else:
            self.result_char = f'{BG.FAIL}X{BG.ENDC}'


import setuptools

setuptools.setup(
    name='testrexx',
    version='0.3.9',
    author='Guilherme Cartier',
    description='Python library for REXX automated testing',
    packages=['trexx',
              'trexx.utilities',
              'trexx.rexx_lib'],
    install_requires=['pyrexx',
                      'mkrexx',
                      'python-decouple',
                      'zowe-zos-files-for-zowe-sdk',
                      'zowe-zos-jobs-for-zowe-sdk'],
    package_data={'': ['*.rex']},
    include_package_data=True,
    python_requires='>=3.6'
)



import os
import logging
from .utilities import TestFile
from .utilities import BG
from .utilities import LOG
from mkrexx import MakeRexx
from decouple import config

INTERNAL_REXX_LIB = 'rexx_lib'
ASSERTION_LIB = os.path.join(os.path.dirname(os.path.abspath(__file__)), INTERNAL_REXX_LIB, 'assertions.rex')

class TRexx():

    def __init__(self):
        self.test_code_path = config('TREXX_TEST_CODE_PATH')
        self.test_lib_path = config('TREXX_TEST_LIB_PATH')
        self.test_files = []
        self.__setup_mkrexx()
        self.__build_tests()
        self.__parse_test_files()

    def __setup_mkrexx(self):
        self.__create_test_build_path()
        self.__load_assertion_lib()

    def __create_test_build_path(self):
        if os.path.isdir(self.test_code_path):
            self.test_build_path = os.path.join(self.test_code_path, 'build')
            try:
                os.mkdir(self.test_build_path)
            except FileExistsError:
                pass
        else:
            # TODO handle this properly
            raise Exception(f'"{self.test_code_path}" is an invalid path and this should be handled properly in the code')    

    def __load_assertion_lib(self):
        with open(ASSERTION_LIB, 'r') as file_object:
            assertion_lib_contents = file_object.read()
        self.assertion_lib_records = [record.rstrip() for record in assertion_lib_contents.split('\n')]

    def __build_tests(self):
        self.mk = MakeRexx(origin_path=self.test_code_path,
                           lib_path=self.test_lib_path,
                           build_path=self.test_build_path)
        self.origin_files = [value for key, value in self.mk.origin_files.items()]
        self.__append_assertion_lib()
        self.mk.build()

    def __append_assertion_lib(self):
        for file in self.origin_files:
            file.code.append_records(self.assertion_lib_records)

    def __parse_test_files(self):
        self.test_files = [TestFile(file_object) for file_name, file_object in self.mk.build_files.items()]

    def __len__(self):
        return len(self.test_files)

    def run(self):
        total_runtime = 0
        LOG.info(f"{BG.OKBLUE}Initiating test run...{BG.ENDC}")
        for test_file in self.test_files:
            test_file.run()
            result_list = [test_case.passed for test_case in test_file.test_definitions]
            LOG.info(f"Display test results...")
            print("\n")
            [test_case.display_test_results() for test_case in test_file.test_definitions]
            total_runtime += test_file._run_time
        print("\n")
        print(f"Total runtime: {BG.OKGREEN}{total_runtime:.3f}{BG.ENDC} secs ⌚")
        if False in result_list:
            print(f"{BG.WARNING}Test completed with failed results.{BG.ENDC}")
            exit(1)



import time
import requests
from zowe.zos_jobs_for_zowe_sdk import Jobs
from .logger import LOG



class TestJcl:

    # TODO this design needs to be improved
    def __init__(self, test_dataset, connection):
        self.zosmf_url = connection['host_url']
        self.zosmf_user = connection['user']
        self.zosmf_password = connection['password']
        self.test_dataset = test_dataset
        self.zos_jobs = Jobs(connection)
        self.base_jcl = f"""//TESTREXX JOB (AUTOMATION),CLASS=A,MSGCLASS=0,         
//             MSGLEVEL=(1,1),REGION=0M,NOTIFY=&SYSUID
//RUNREXX  EXEC PGM=IKJEFT01
//SYSEXEC  DD DSN={self.test_dataset},DISP=SHR
//SYSTSPRT DD SYSOUT=*
//SYSTSIN  DD *"""

    def trigger_rexx(self, rexx_module):
        LOG.debug(f'Triggering REXX module "{rexx_module}"')
        self.base_jcl += f"\n {rexx_module} \n/*"
        self.__run_jcl()

    def __run_jcl(self):
        LOG.info(f'Submitting test job')
        self.jcl_request = self.zos_jobs.submit_plaintext(self.base_jcl)
        self.jobname = self.jcl_request['jobname']
        self.jcl_status = self.jcl_request['status']
        self.retcode = self.jcl_request['retcode']
        self.jobid = self.jcl_request['jobid']
        self.files_url = self.jcl_request['files-url']
        self.__wait_for_job_to_complete()
        self.result_records = self.__get_result_records()

    def __wait_for_job_to_complete(self):
        while self.retcode != 'CC 0000':
            time.sleep(1.5)
            self.retcode = self.zos_jobs.get_job_status(self.jobname, self.jobid)['retcode']
            LOG.debug(f'Job: {self.jobname} ({self.jobid}) - Return code: {self.retcode}')

    # TODO dependencies on requests module are not ideal (need to improve zowe sdk)
    def __get_result_records(self):
        LOG.info(f"Retrieving output from job {self.jobname} ({self.jobid})")
        file_list = requests.get(self.files_url, auth=(self.zosmf_user, self.zosmf_password), verify=False)
        for item in file_list.json():
            if item['ddname'] == 'SYSTSPRT':
                job_result_object = requests.get(item['records-url'], auth=(self.zosmf_user, self.zosmf_password), verify=False)
                return job_result_object.text.split("\n")
        LOG.error(f"Unable to find valid output for job {self.jobname} ({self.jobid})")
        return []

    # TODO besides removing the dependency on requests I also need to do error checking
    def delete_from_spool(self):
        LOG.debug(f"Deleting {self.jobname}({self.jobid}) from spool")
        requests.delete(f'https://{self.zosmf_url}/zosmf/restjobs/jobs/{self.jobname}/{self.jobid}', auth=(self.zosmf_user, self.zosmf_password), verify=False)



from .exceptions import ZoweConfigurationNotFound
from .test_case import TestCase
from .background_colors import BG
from .test_jcl import TestJcl
from .test_file import TestFile
from .logger import LOG